@@ -1,44 +1,49 @@
+# -*- coding: utf-8 -*-
+
 import io
+import json
 import logging
 import socket
 import sys
 import time
 import warnings
-import pytest
+from test import LONG_TIMEOUT, SHORT_TIMEOUT, onlyPy2
+from threading import Event
 
 import mock
+import pytest
+import six
 
-from .. import TARPIT_HOST, VALID_SOURCE_ADDRESSES, INVALID_SOURCE_ADDRESSES
-from ..port_helpers import find_unused_port
-from urllib3 import encode_multipart_formdata, HTTPConnectionPool
+from dummyserver.server import HAS_IPV6_AND_DNS, NoIPv6Warning
+from dummyserver.testcase import HTTPDummyServerTestCase, SocketDummyServerTestCase
+from urllib3 import HTTPConnectionPool, encode_multipart_formdata
+from urllib3._collections import HTTPHeaderDict
+from urllib3.connection import _get_default_user_agent
 from urllib3.exceptions import (
     ConnectTimeoutError,
-    EmptyPoolError,
     DecodeError,
+    EmptyPoolError,
     MaxRetryError,
-    ReadTimeoutError,
     NewConnectionError,
+    ReadTimeoutError,
     UnrewindableBodyError,
 )
 from urllib3.packages.six import b, u
 from urllib3.packages.six.moves.urllib.parse import urlencode
-from urllib3.util.retry import Retry, RequestHistory
+from urllib3.util import SKIP_HEADER, SKIPPABLE_HEADERS
+from urllib3.util.retry import RequestHistory, Retry
 from urllib3.util.timeout import Timeout
 
-from dummyserver.testcase import HTTPDummyServerTestCase, SocketDummyServerTestCase
-from dummyserver.server import NoIPv6Warning, HAS_IPV6_AND_DNS
+from .. import INVALID_SOURCE_ADDRESSES, TARPIT_HOST, VALID_SOURCE_ADDRESSES
+from ..port_helpers import find_unused_port
 
-from threading import Event
+pytestmark = pytest.mark.flaky
 
 log = logging.getLogger("urllib3.connectionpool")
 log.setLevel(logging.NOTSET)
 log.addHandler(logging.StreamHandler(sys.stdout))
 
 
-SHORT_TIMEOUT = 0.001
-LONG_TIMEOUT = 0.03
-
-
 def wait_for_socket(ready_event):
     ready_event.wait()
     ready_event.clear()
@@ -49,19 +54,16 @@ class TestConnectionPoolTimeouts(SocketDummyServerTestCase):
         block_event = Event()
         ready_event = self.start_basic_handler(block_send=block_event, num=2)
 
-        # Pool-global timeout
-        with HTTPConnectionPool(
-            self.host, self.port, timeout=SHORT_TIMEOUT, retries=False
-        ) as pool:
+        with HTTPConnectionPool(self.host, self.port, retries=False) as pool:
             wait_for_socket(ready_event)
             with pytest.raises(ReadTimeoutError):
-                pool.request("GET", "/")
+                pool.request("GET", "/", timeout=SHORT_TIMEOUT)
             block_event.set()  # Release block
 
             # Shouldn't raise this time
             wait_for_socket(ready_event)
             block_event.set()  # Pre-release block
-            pool.request("GET", "/")
+            pool.request("GET", "/", timeout=LONG_TIMEOUT)
 
     def test_conn_closed(self):
         block_event = Event()
@@ -73,9 +75,8 @@ class TestConnectionPoolTimeouts(SocketDummyServerTestCase):
             conn = pool._get_conn()
             pool._put_conn(conn)
             try:
-                pool.urlopen("GET", "/")
-                self.fail("The request should fail with a timeout error.")
-            except ReadTimeoutError:
+                with pytest.raises(ReadTimeoutError):
+                    pool.urlopen("GET", "/")
                 if conn.sock:
                     with pytest.raises(socket.error):
                         conn.sock.recv(1024)
@@ -87,21 +88,13 @@ class TestConnectionPoolTimeouts(SocketDummyServerTestCase):
     def test_timeout(self):
         # Requests should time out when expected
         block_event = Event()
-        ready_event = self.start_basic_handler(block_send=block_event, num=6)
+        ready_event = self.start_basic_handler(block_send=block_event, num=3)
 
         # Pool-global timeout
-        timeout = Timeout(read=SHORT_TIMEOUT)
+        short_timeout = Timeout(read=SHORT_TIMEOUT)
         with HTTPConnectionPool(
-            self.host, self.port, timeout=timeout, retries=False
+            self.host, self.port, timeout=short_timeout, retries=False
         ) as pool:
-
-            wait_for_socket(ready_event)
-            conn = pool._get_conn()
-            with pytest.raises(ReadTimeoutError):
-                pool._make_request(conn, "GET", "/")
-            pool._put_conn(conn)
-            block_event.set()  # Release request
-
             wait_for_socket(ready_event)
             block_event.clear()
             with pytest.raises(ReadTimeoutError):
@@ -110,45 +103,24 @@ class TestConnectionPoolTimeouts(SocketDummyServerTestCase):
 
         # Request-specific timeouts should raise errors
         with HTTPConnectionPool(
-            self.host, self.port, timeout=LONG_TIMEOUT, retries=False
+            self.host, self.port, timeout=short_timeout, retries=False
         ) as pool:
-
-            conn = pool._get_conn()
             wait_for_socket(ready_event)
             now = time.time()
             with pytest.raises(ReadTimeoutError):
-                pool._make_request(conn, "GET", "/", timeout=timeout)
-            delta = time.time() - now
-            block_event.set()  # Release request
-
-            message = "timeout was pool-level LONG_TIMEOUT rather than request-level SHORT_TIMEOUT"
-            assert delta < LONG_TIMEOUT, message
-            pool._put_conn(conn)
-
-            wait_for_socket(ready_event)
-            now = time.time()
-            with pytest.raises(ReadTimeoutError):
-                pool.request("GET", "/", timeout=timeout)
+                pool.request("GET", "/", timeout=LONG_TIMEOUT)
             delta = time.time() - now
 
-            message = "timeout was pool-level LONG_TIMEOUT rather than request-level SHORT_TIMEOUT"
-            assert delta < LONG_TIMEOUT, message
+            message = "timeout was pool-level SHORT_TIMEOUT rather than request-level LONG_TIMEOUT"
+            assert delta >= LONG_TIMEOUT, message
             block_event.set()  # Release request
 
-            # Timeout int/float passed directly to request and _make_request should
-            # raise a request timeout
+            # Timeout passed directly to request should raise a request timeout
             wait_for_socket(ready_event)
             with pytest.raises(ReadTimeoutError):
                 pool.request("GET", "/", timeout=SHORT_TIMEOUT)
             block_event.set()  # Release request
 
-            wait_for_socket(ready_event)
-            conn = pool._new_conn()
-            # FIXME: This assert flakes sometimes. Not sure why.
-            with pytest.raises(ReadTimeoutError):
-                pool._make_request(conn, "GET", "/", timeout=SHORT_TIMEOUT)
-            block_event.set()  # Release request
-
     def test_connect_timeout(self):
         url = "/"
         host, port = TARPIT_HOST, 80
@@ -385,11 +357,9 @@ class TestConnectionPool(HTTPDummyServerTestCase):
         """ ECONNREFUSED error should raise a connection error, with retries """
         port = find_unused_port()
         with HTTPConnectionPool(self.host, port) as pool:
-            try:
+            with pytest.raises(MaxRetryError) as e:
                 pool.request("GET", "/", retries=Retry(connect=3))
-                self.fail("Should have failed with a connection error.")
-            except MaxRetryError as e:
-                assert type(e.reason) == NewConnectionError
+            assert type(e.value.reason) == NewConnectionError
 
     def test_timeout_success(self):
         timeout = Timeout(connect=3, read=5, total=None)
@@ -441,11 +411,9 @@ class TestConnectionPool(HTTPDummyServerTestCase):
 
     def test_bad_connect(self):
         with HTTPConnectionPool("badhost.invalid", self.port) as pool:
-            try:
+            with pytest.raises(MaxRetryError) as e:
                 pool.request("GET", "/", retries=5)
-                self.fail("should raise timeout exception here")
-            except MaxRetryError as e:
-                assert type(e.reason) == NewConnectionError
+            assert type(e.value.reason) == NewConnectionError
 
     def test_keepalive(self):
         with HTTPConnectionPool(self.host, self.port, block=True, maxsize=1) as pool:
@@ -460,7 +428,6 @@ class TestConnectionPool(HTTPDummyServerTestCase):
         with HTTPConnectionPool(
             self.host, self.port, block=True, maxsize=1, timeout=2
         ) as pool:
-
             r = pool.request(
                 "GET", "/keepalive?close=1", retries=0, headers={"Connection": "close"}
             )
@@ -624,7 +591,6 @@ class TestConnectionPool(HTTPDummyServerTestCase):
         with HTTPConnectionPool(
             self.host, self.port, block=True, maxsize=1, timeout=2
         ) as pool:
-
             payload_size = 1024 * 2
             first_chunk = 512
 
@@ -717,6 +683,12 @@ class TestConnectionPool(HTTPDummyServerTestCase):
             with pytest.raises(MaxRetryError):
                 pool.request("GET", "/test", retries=2)
 
+    @pytest.mark.parametrize("char", [" ", "\r", "\n", "\x00"])
+    def test_invalid_method_not_allowed(self, char):
+        with pytest.raises(ValueError):
+            with HTTPConnectionPool(self.host, self.port) as pool:
+                pool.request("GET" + char, "/")
+
     def test_percent_encode_invalid_target_chars(self):
         with HTTPConnectionPool(self.host, self.port) as pool:
             r = pool.request("GET", "/echo_params?q=\r&k=\n \n")
@@ -821,17 +793,236 @@ class TestConnectionPool(HTTPDummyServerTestCase):
             response = pool.request("GET", "http://LoCaLhOsT:%d/" % self.port)
             assert response.status == 200
 
+    def test_preserves_path_dot_segments(self):
+        """ ConnectionPool preserves dot segments in the URI """
+        with HTTPConnectionPool(self.host, self.port) as pool:
+            response = pool.request("GET", "/echo_uri/seg0/../seg2")
+            assert response.data == b"/echo_uri/seg0/../seg2"
+
+    def test_default_user_agent_header(self):
+        """ ConnectionPool has a default user agent """
+        default_ua = _get_default_user_agent()
+        custom_ua = "I'm not a web scraper, what are you talking about?"
+        custom_ua2 = "Yet Another User Agent"
+        with HTTPConnectionPool(self.host, self.port) as pool:
+            # Use default user agent if no user agent was specified.
+            r = pool.request("GET", "/headers")
+            request_headers = json.loads(r.data.decode("utf8"))
+            assert request_headers.get("User-Agent") == _get_default_user_agent()
+
+            # Prefer the request user agent over the default.
+            headers = {"UsEr-AGENt": custom_ua}
+            r = pool.request("GET", "/headers", headers=headers)
+            request_headers = json.loads(r.data.decode("utf8"))
+            assert request_headers.get("User-Agent") == custom_ua
+
+            # Do not modify pool headers when using the default user agent.
+            pool_headers = {"foo": "bar"}
+            pool.headers = pool_headers
+            r = pool.request("GET", "/headers")
+            request_headers = json.loads(r.data.decode("utf8"))
+            assert request_headers.get("User-Agent") == default_ua
+            assert "User-Agent" not in pool_headers
+
+            pool.headers.update({"User-Agent": custom_ua2})
+            r = pool.request("GET", "/headers")
+            request_headers = json.loads(r.data.decode("utf8"))
+            assert request_headers.get("User-Agent") == custom_ua2
+
+    @pytest.mark.parametrize(
+        "headers",
+        [
+            None,
+            {},
+            {"User-Agent": "key"},
+            {"user-agent": "key"},
+            {b"uSeR-AgEnT": b"key"},
+            {b"user-agent": "key"},
+        ],
+    )
+    @pytest.mark.parametrize("chunked", [True, False])
+    def test_user_agent_header_not_sent_twice(self, headers, chunked):
+        with HTTPConnectionPool(self.host, self.port) as pool:
+            r = pool.request("GET", "/headers", headers=headers, chunked=chunked)
+            request_headers = json.loads(r.data.decode("utf8"))
+
+            if not headers:
+                assert request_headers["User-Agent"].startswith("python-urllib3/")
+                assert "key" not in request_headers["User-Agent"]
+            else:
+                assert request_headers["User-Agent"] == "key"
+
+    def test_no_user_agent_header(self):
+        """ ConnectionPool can suppress sending a user agent header """
+        custom_ua = "I'm not a web scraper, what are you talking about?"
+        with HTTPConnectionPool(self.host, self.port) as pool:
+            # Suppress user agent in the request headers.
+            no_ua_headers = {"User-Agent": SKIP_HEADER}
+            r = pool.request("GET", "/headers", headers=no_ua_headers)
+            request_headers = json.loads(r.data.decode("utf8"))
+            assert "User-Agent" not in request_headers
+            assert no_ua_headers["User-Agent"] == SKIP_HEADER
+
+            # Suppress user agent in the pool headers.
+            pool.headers = no_ua_headers
+            r = pool.request("GET", "/headers")
+            request_headers = json.loads(r.data.decode("utf8"))
+            assert "User-Agent" not in request_headers
+            assert no_ua_headers["User-Agent"] == SKIP_HEADER
+
+            # Request headers override pool headers.
+            pool_headers = {"User-Agent": custom_ua}
+            pool.headers = pool_headers
+            r = pool.request("GET", "/headers", headers=no_ua_headers)
+            request_headers = json.loads(r.data.decode("utf8"))
+            assert "User-Agent" not in request_headers
+            assert no_ua_headers["User-Agent"] == SKIP_HEADER
+            assert pool_headers.get("User-Agent") == custom_ua
+
+    @pytest.mark.parametrize(
+        "accept_encoding",
+        [
+            "Accept-Encoding",
+            "accept-encoding",
+            b"Accept-Encoding",
+            b"accept-encoding",
+            None,
+        ],
+    )
+    @pytest.mark.parametrize("host", ["Host", "host", b"Host", b"host", None])
+    @pytest.mark.parametrize(
+        "user_agent", ["User-Agent", "user-agent", b"User-Agent", b"user-agent", None]
+    )
+    @pytest.mark.parametrize("chunked", [True, False])
+    def test_skip_header(self, accept_encoding, host, user_agent, chunked):
+        headers = {}
+
+        if accept_encoding is not None:
+            headers[accept_encoding] = SKIP_HEADER
+        if host is not None:
+            headers[host] = SKIP_HEADER
+        if user_agent is not None:
+            headers[user_agent] = SKIP_HEADER
+
+        with HTTPConnectionPool(self.host, self.port) as pool:
+            r = pool.request("GET", "/headers", headers=headers, chunked=chunked)
+        request_headers = json.loads(r.data.decode("utf8"))
+
+        if accept_encoding is None:
+            assert "Accept-Encoding" in request_headers
+        else:
+            assert accept_encoding not in request_headers
+        if host is None:
+            assert "Host" in request_headers
+        else:
+            assert host not in request_headers
+        if user_agent is None:
+            assert "User-Agent" in request_headers
+        else:
+            assert user_agent not in request_headers
+
+    @pytest.mark.parametrize("header", ["Content-Length", "content-length"])
+    @pytest.mark.parametrize("chunked", [True, False])
+    def test_skip_header_non_supported(self, header, chunked):
+        with HTTPConnectionPool(self.host, self.port) as pool:
+            with pytest.raises(ValueError) as e:
+                pool.request(
+                    "GET", "/headers", headers={header: SKIP_HEADER}, chunked=chunked
+                )
+            assert (
+                str(e.value)
+                == "urllib3.util.SKIP_HEADER only supports 'Accept-Encoding', 'Host', 'User-Agent'"
+            )
+
+            # Ensure that the error message stays up to date with 'SKIP_HEADER_SUPPORTED_HEADERS'
+            assert all(
+                ("'" + header.title() + "'") in str(e.value)
+                for header in SKIPPABLE_HEADERS
+            )
+
+    @pytest.mark.parametrize("chunked", [True, False])
+    @pytest.mark.parametrize("pool_request", [True, False])
+    @pytest.mark.parametrize("header_type", [dict, HTTPHeaderDict])
+    def test_headers_not_modified_by_request(self, chunked, pool_request, header_type):
+        # Test that the .request*() methods of ConnectionPool and HTTPConnection
+        # don't modify the given 'headers' structure, instead they should
+        # make their own internal copies at request time.
+        headers = header_type()
+        headers["key"] = "val"
+
+        with HTTPConnectionPool(self.host, self.port) as pool:
+            pool.headers = headers
+            if pool_request:
+                pool.request("GET", "/headers", chunked=chunked)
+            else:
+                conn = pool._get_conn()
+                if chunked:
+                    conn.request_chunked("GET", "/headers")
+                else:
+                    conn.request("GET", "/headers")
+
+            assert pool.headers == {"key": "val"}
+            assert isinstance(pool.headers, header_type)
+
+        with HTTPConnectionPool(self.host, self.port) as pool:
+            if pool_request:
+                pool.request("GET", "/headers", headers=headers, chunked=chunked)
+            else:
+                conn = pool._get_conn()
+                if chunked:
+                    conn.request_chunked("GET", "/headers", headers=headers)
+                else:
+                    conn.request("GET", "/headers", headers=headers)
+
+            assert headers == {"key": "val"}
+
+    def test_bytes_header(self):
+        with HTTPConnectionPool(self.host, self.port) as pool:
+            headers = {"User-Agent": b"test header"}
+            r = pool.request("GET", "/headers", headers=headers)
+            request_headers = json.loads(r.data.decode("utf8"))
+            assert "User-Agent" in request_headers
+            assert request_headers["User-Agent"] == "test header"
+
+    @pytest.mark.parametrize(
+        "user_agent", [u"Schönefeld/1.18.0", u"Schönefeld/1.18.0".encode("iso-8859-1")]
+    )
+    def test_user_agent_non_ascii_user_agent(self, user_agent):
+        if six.PY2 and not isinstance(user_agent, str):
+            pytest.skip(
+                "Python 2 raises UnicodeEncodeError when passed a unicode header"
+            )
+
+        with HTTPConnectionPool(self.host, self.port, retries=False) as pool:
+            r = pool.urlopen(
+                "GET",
+                "/headers",
+                headers={"User-Agent": user_agent},
+            )
+            request_headers = json.loads(r.data.decode("utf8"))
+            assert "User-Agent" in request_headers
+            assert request_headers["User-Agent"] == u"Schönefeld/1.18.0"
+
+    @onlyPy2
+    def test_user_agent_non_ascii_fails_on_python_2(self):
+        with HTTPConnectionPool(self.host, self.port, retries=False) as pool:
+            with pytest.raises(UnicodeEncodeError) as e:
+                pool.urlopen(
+                    "GET",
+                    "/headers",
+                    headers={"User-Agent": u"Schönefeld/1.18.0"},
+                )
+            assert str(e.value) == (
+                "'ascii' codec can't encode character u'\\xf6' in "
+                "position 3: ordinal not in range(128)"
+            )
+
 
 class TestRetry(HTTPDummyServerTestCase):
     def test_max_retry(self):
         with HTTPConnectionPool(self.host, self.port) as pool:
-            try:
-                r = pool.request("GET", "/redirect", fields={"target": "/"}, retries=0)
-                self.fail(
-                    "Failed to raise MaxRetryError exception, returned %r" % r.status
-                )
-            except MaxRetryError:
-                pass
+            with pytest.raises(MaxRetryError):
+                pool.request("GET", "/redirect", fields={"target": "/"}, retries=0)
 
     def test_disabled_retry(self):
         """ Disabled retries should disable redirect handling. """
@@ -1135,11 +1326,9 @@ class TestFileBodiesOnRetryOrRedirect(HTTPDummyServerTestCase):
         # which is unsupported by BytesIO.
         headers = {"Content-Length": "8"}
         with HTTPConnectionPool(self.host, self.port, timeout=0.1) as pool:
-            try:
+            with pytest.raises(UnrewindableBodyError) as e:
                 pool.urlopen("PUT", url, headers=headers, body=body)
-                self.fail("PUT successful despite failed rewind.")
-            except UnrewindableBodyError as e:
-                assert "Unable to record file position for" in str(e)
+            assert "Unable to record file position for" in str(e.value)
 
 
 class TestRetryPoolSize(HTTPDummyServerTestCase):
